/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "tMovieView.h"

@implementation tMovieView

- (void) awakeFromNib
{
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(closeWindow:)  name:@"movieWindowCloseNotification" object:nil];
}

- (void) closeWindow:(NSNotification *)note
{
    //NSLog(@"close");
    [[self window] close];
}

- (BOOL)performKeyEquivalent:(NSEvent *)theEvent
{
    BOOL result;
    NSString *chars;
    unichar uch;
    
    result = [super performKeyEquivalent:theEvent];
    chars = [theEvent characters];
    uch = [chars characterAtIndex:0];
    if (uch == 'f') {
        [tPlayback fullScreen];
        return YES;
    }
    if (uch == '\t') {
        if ([theEvent modifierFlags] & NSAlternateKeyMask)
            [tPlayback lastMovie];
        else
            [tPlayback nextMovie];
        return YES;
    }
    if (uch == NSUpArrowFunctionKey || uch == NSDownArrowFunctionKey) {
        [tPlayback performSelector:@selector(refreshVolume:) withObject:nil afterDelay:0];
    }
    if ([theEvent modifierFlags] & NSCommandKeyMask ||
        [theEvent modifierFlags] & NSControlKeyMask) // ||
        //[theEvent modifierFlags] & NSAlternateKeyMask)
            return NO;
    BOOL forward = NO;
    switch (uch) {
        case NSRightArrowFunctionKey: forward = YES;
        case NSLeftArrowFunctionKey:
            if (doubles && [theEvent modifierFlags] & NSAlternateKeyMask) {
                MHMovie *movie = (MHMovie *)[self movie];
                if ([movie isKindOfClass:[MHMovie class]]) {
                    double seconds = [movie currentTimeIndexSeconds];
                    if (forward) {
                        NSString *lang = nil;
                        if ([[movie chapters] multilingual]) lang = [movie currentLanguage];
                        seconds = [[movie chapters] nextTimeForTime:seconds language:lang];
                        if (seconds == -1) seconds = [movie durationSeconds] - 0.5;
                    } else
                        seconds = [[movie chapters] prevTimeForTime:seconds];
                    [movie setCurrentTimeIndexSeconds:seconds];
                } else return NO;
            } else if (doubles) {
                if (forward) [self stepForward:nil];
                else [self stepBack:nil];
            }
            doubles = !doubles;
            return YES;
        case 'f':
            [tPlayback fullScreen];
            return YES;
    }
    return NO;
}

- (void)selectAll:(id)sender
{
    if ([sender respondsToSelector:@selector(title)] && [[sender title] hasSuffix:@"Controller"])
	[tPlayback toggleController:nil];
    else
	[super selectAll:sender];
}

- (void)mouseDown:(NSEvent *)event
{
    if ([event modifierFlags] & NSControlKeyMask)
        [self rightMouseDown:event];
    else [super mouseDown:event];
}

- (void)rightMouseDown:(NSEvent *)theEvent
{
    NSMenu *menu = [[[NSMenu alloc] init] autorelease];
    NSMenuItem *item = [[[NSMenuItem alloc] initWithTitle:@"Play" action:@selector(start:) keyEquivalent:@""] autorelease];
    [item setTarget:self];
    [menu addItem:item];
    item = [[[NSMenuItem alloc] initWithTitle:@"Stop" action:@selector(stop:) keyEquivalent:@""] autorelease];
    [item setTarget:self];
    [menu addItem:item];
    NSString *title = @"Show Controller";
    if ([self isControllerVisible])
	title = @"Hide Controller";
    item = [[[NSMenuItem alloc] initWithTitle:title action:@selector(selectAll:) keyEquivalent:@""] autorelease];
    [item setTarget:self];
    [menu addItem:item];
    if ([[self movie] respondsToSelector:@selector(chapters)] && [(MHChapters *)[(MHMovie *)[self movie] chapters] count]) {
	NSMenu *submenu = [[[NSMenu alloc] init] autorelease];
	item = [[[NSMenuItem alloc] initWithTitle:@"Chapters" action:nil keyEquivalent:@""] autorelease];
	[item setSubmenu:submenu];
	[menu addItem:item];
	NSString *lang = nil;
	if ([[(MHMovie *)[self movie] chapters] multilingual]) lang = [(MHMovie *)[self movie] currentLanguage];
	NSArray *names = [[(MHMovie *)[self movie] chapters] chapterNamesForLanguage:lang];
	int i, n = [names count];
	for (i=0;i<n;i++) {
	    item = [[[NSMenuItem alloc] initWithTitle:[names objectAtIndex:i] action:@selector(gotoChapter:) keyEquivalent:@""] autorelease];
	    [item setTarget:[tPlayback nibInstance]];
	    [submenu addItem:item];
	}
    }
    
    [NSMenu popUpContextMenu:menu withEvent:theEvent forView:self];
}

@end
